/*
LibHFZ, a library to load and save HF2/HFZ compressed heightfield files.

Copyright (C) 2007 Aaron Torpy

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <memory.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
	

#include "libhfz.h"
#include "zlib\include\zlib.h"
#include "zlib\include\zconf.h"


//
//
// High-level HFZ functions
//
//


long hfzSave(const char* lpFileName, long Format, float* pData, long nx, long ny, int TileSize, float Precis, float HorizScale, char* lpExtHeader, long ExtHeaderLength) {

	// check format
	switch(Format) {
	default:
		return ERROR_INVALID_FORMAT;
	case FORMAT_HF2:
	case FORMAT_HF2_GZ:
		break;
	}

	// check settings
	if(Precis<=0) {
		return ERROR_INVALID_PRECIS;
	}
	if(TileSize<8 || TileSize>65535) {
		return ERROR_INVALID_TILE_SIZE;
	}
	if(nx<=0 || ny<=0) {
		return ERROR_INVALID_MAP_SIZE;
	}
	
	// determine number of blocks
	int nBlocksX = nx / TileSize;
	int nBlocksY = ny / TileSize;
	if(nx%TileSize)	nBlocksX++;
	if(ny%TileSize)	nBlocksY++;
	int BlockCount = nBlocksX * nBlocksY;

	// open file for writing
	hfzFile* fs;
	if(!(fs = hfzOpen(lpFileName, Format, "wb"))) {
		return ERROR_OPEN_FAILED;
	}

	// making file header
	hfzHeader FH_HFZ;
	FH_HFZ.FileVersionNo = 0; // this is the very first implementation
	FH_HFZ.nx = nx;
	FH_HFZ.ny = ny;
	FH_HFZ.TileSize = TileSize;
	FH_HFZ.Precis = Precis;
	FH_HFZ.HorizScale = HorizScale;
	FH_HFZ.ExtHeaderLength = ExtHeaderLength;
	FH_HFZ.pExtHeaderData = lpExtHeader;

	// write header
	if(ERROR_OK!=hfzWriteHeader(fs, FH_HFZ)) {
		hfzClose(fs);
		return ERROR_WRITE_FAILED;
	}

	// writing tiles
	int rval = ERROR_OK;
	for(int bj = 0; bj<nBlocksY && rval==ERROR_OK; bj++) {
		for(int bi = 0; bi<nBlocksX && rval==ERROR_OK; bi++) {
			rval = hfzWriteTile(fs, FH_HFZ, bi, bj, pData);
		}
	}

	// close file
	hfzClose(fs);

	return rval;
}

long hfzLoad(const char* lpFileName, hfzHeader& fh, float *pData) {

	if(!pData)
		return ERROR_INVALID_HANDLE;

	hfzResetHeader(fh);

	int rval;

	hfzFile* fs;
	fs = hfzOpen(lpFileName, FORMAT_HF2_GZ, "rb"); // always treat as gzip; zlib handles raw too
	if(!fs) {
		return ERROR_OPEN_FAILED;
	}

	rval = hfzReadHeader(fs, fh);
	if(ERROR_OK!=rval) {
		hfzClose(fs);
		return rval;
	}

	// shorthand (shorter hand, anyhow)
	int nx = fh.nx;
	int ny = fh.ny;
	int TileSize = fh.TileSize;

	// debugging
	//TempStr.Format("VersionNo = %d\r\nnx = %d\r\nny = %d\r\nBlockSize = %d\r\nPrecis = %.2f\r\nHorizScale = %.2f\r\nExtHeaderLength = %d", FH_HFZ.FileVersionNo, FH_HFZ.nx, FH_HFZ.ny, FH_HFZ.BlockSize, FH_HFZ.Precis, FH_HFZ.HorizScale, FH_HFZ.ExtHeaderLength);
	//AfxMessageBox(TempStr);

	// determine number of blocks
	int nBlocksX = nx / TileSize;
	int nBlocksY = ny / TileSize;
	if(nx%TileSize)	nBlocksX++;
	if(ny%TileSize)	nBlocksY++;
	int BlockCount = nBlocksX * nBlocksY;

	// start reading blocks
	int bnx, bny;
	
	bool ErrorFlag = false;

	char c;
	short s;
	long li;

	rval = ERROR_OK;
	for(int bj = 0; bj<nBlocksY && rval == ERROR_OK; bj++) {
		for(int bi = 0; bi<nBlocksX && rval == ERROR_OK; bi++) {

			rval = hfzReadTile(fs, fh, bi, bj, pData);
		}
	}

	hfzClose(fs);

	return rval;
}

long hfzLoadEx(const char* lpFileName, hfzHeader& fh, float **h_pData) {

	hfzResetHeader(fh);
	int rval = hfzReadHeader2(lpFileName, fh);
	if(ERROR_OK!=rval) {
        return rval;		
	}

	// allocate buffer
	float* pData = (float*)calloc(fh.nx * fh.ny, 4);
	if(!pData) {
		hfzResetHeader(fh);
		return ERROR_ALLOC_FAILED;
	}

	// load 
	rval = hfzLoad(lpFileName, fh, pData);
	if(ERROR_OK!=rval) {
		hfzResetHeader(fh);
		free(pData);
		return rval;
	}

	*h_pData = pData;
	return ERROR_OK;
}

//
//
// Medium-level HFZ functions
//
//

long hfzWriteHeader(hfzFile* fs, hfzHeader& fh) {

	// copy header into buffer
	char HeaderBuf[28];
	int HeaderBufPos = 0;
	sprintf(HeaderBuf+HeaderBufPos, "HF2"); HeaderBufPos+=4;
	memcpy(HeaderBuf+HeaderBufPos, &(fh.FileVersionNo), 2); HeaderBufPos+=2;
	memcpy(HeaderBuf+HeaderBufPos, &(fh.nx), 4); HeaderBufPos+=4;
	memcpy(HeaderBuf+HeaderBufPos, &(fh.ny), 4); HeaderBufPos+=4;
	memcpy(HeaderBuf+HeaderBufPos, &(fh.TileSize), 2); HeaderBufPos+=2;
	memcpy(HeaderBuf+HeaderBufPos, &(fh.Precis), 4); HeaderBufPos+=4;
	memcpy(HeaderBuf+HeaderBufPos, &(fh.HorizScale), 4); HeaderBufPos+=4;
	memcpy(HeaderBuf+HeaderBufPos, &(fh.ExtHeaderLength), 4); HeaderBufPos+=4;

	// write header
	if(28!=hfzWrite(fs, HeaderBuf, 28)) 
		return ERROR_WRITE_HEADER_FAILED;

	// write extended header
	if(fh.ExtHeaderLength!=hfzWrite(fs, fh.pExtHeaderData, fh.ExtHeaderLength))
		return ERROR_WRITE_EXTHEAD_FAILED;

	return ERROR_OK;
}

long hfzReadHeader(hfzFile* fs, hfzHeader& fh) {

	hfzResetHeader(fh);

	// read the core header
	char HeaderBuf[28];
	if(28!=hfzRead(fs, HeaderBuf, 28)) {
		return ERROR_READ_FAILED;
	}

	// check for "HF2\0" in leading 4 bytes
	HeaderBuf[3] = 0;
	if(strcmp(HeaderBuf, "HF2")) {
		return ERROR_INVALID_FORMAT;
	}

	int HeaderPos = 4;
	memcpy(&(fh.FileVersionNo), HeaderBuf+HeaderPos, 2); HeaderPos+=2;
	memcpy(&(fh.nx), HeaderBuf+HeaderPos, 4); HeaderPos+=4;
	memcpy(&(fh.ny), HeaderBuf+HeaderPos, 4); HeaderPos+=4;
	memcpy(&(fh.TileSize), HeaderBuf+HeaderPos, 2); HeaderPos+=2;
	memcpy(&(fh.Precis), HeaderBuf+HeaderPos, 4); HeaderPos+=4;
	memcpy(&(fh.HorizScale), HeaderBuf+HeaderPos, 4); HeaderPos+=4;
	memcpy(&(fh.ExtHeaderLength), HeaderBuf+HeaderPos, 4);
	
	unsigned long hl = fh.ExtHeaderLength;
    if(hl) {
		if(!(fh.pExtHeaderData = (char*)calloc(hl, 1))) {
			return ERROR_ALLOC_FAILED;
		}

		if(hl!=hfzRead(fs, fh.pExtHeaderData, hl)) {
			return ERROR_READ_FAILED;
		}
	}

	return ERROR_OK;
}

long hfzReadHeader2(const char* lpFileName, hfzHeader& fh) {
	hfzFile* fs;
	if(!(fs=hfzOpen(lpFileName, FORMAT_HF2_GZ, "rb"))) {
		return ERROR_OPEN_FAILED;
	}

	if(ERROR_OK!=hfzReadHeader(fs, fh)) {
		return ERROR_READ_FAILED;
	}

	hfzClose(fs);

	return ERROR_OK;
}


long hfzReadTile(hfzFile* fs, hfzHeader& fh, long TileX, long TileY, float* pMapData) {

	int i1, i2, j1, j2, i, j;

	char c;
	short s;
	long li;

	int TileSize = fh.TileSize;
	int nx = fh.nx;
	int ny = fh.ny;

	i1 = TileX * TileSize;
	j1 = TileY * TileSize;
	i2 = i1 + TileSize;
	j2 = j1 + TileSize;

	if(i2>=nx)	i2 = nx;
	if(j2>=ny)	j2 = ny;

	int bnx = i2 - i1;
	int bny = j2 - j1;

	// read vert offset and sale
	float VertOffset;
	float VertScale;
	if(4!=hfzRead(fs, &VertScale, 4))
		return ERROR_READ_FAILED;
	if(4!=hfzRead(fs, &VertOffset, 4))
		return ERROR_READ_FAILED;

	// debugging
	//TempStr.Format("VertScale = %f\r\nVertOffset = %f", VertOffset, VertScale);
	//AfxMessageBox(TempStr);

	for(j=j1; j<j2; j++) {

		char LineDepth; // 1, 2, or 4
		int FirstVal;

		if(1!=hfzRead(fs, &LineDepth, 1))
			return ERROR_READ_FAILED;
		if(4!=hfzRead(fs, &FirstVal, 4))
			return ERROR_READ_FAILED;

		float f = (float)FirstVal * VertScale + VertOffset;

		// set first pixel
		pMapData[i1 + nx * j] = f;

		long LastVal = FirstVal;

		for(i=i1+1; i<i2; i++) {
			switch(LineDepth) {
			case 1:
				if(1!=hfzRead(fs, &c, LineDepth))	return ERROR_READ_FAILED;
				li = (long)c;
				break;
			case 2:
				if(2!=hfzRead(fs, &s, LineDepth))	return ERROR_READ_FAILED;
				li = (long)s;
				break;
			case 4:
				if(4!=hfzRead(fs, &li, LineDepth))	return ERROR_READ_FAILED;
				break;
			}

			f = (float)(li +  LastVal) * VertScale + VertOffset;
			LastVal = li + LastVal;

			pMapData[i + nx * j] = f;
		}
	}
	
	return ERROR_OK;
}

long hfzWriteTile(hfzFile* fs, hfzHeader& fh, long TileX, long TileY, float* pMapData) {

	if(!fs)
		return ERROR_INVALID_HANDLE;

	if(!pMapData)
		return ERROR_INVALID_HANDLE;

	int i, j;
	int TempInt;
	float f;
	float HFmin, HFmax;
	char c;
	short s;
	int i1, j1, i2, j2, bnx, bny;
	int FirstVal;
	float VertScale;
	float VertOffset;

	long nx = fh.nx;
	long ny = fh.ny;
	long TileSize = fh.TileSize;
	float Precis = fh.Precis;


	i1 = TileX * TileSize;
	j1 = TileY * TileSize;
	i2 = i1 + TileSize;
	j2 = j1 + TileSize;

	if(i2>=nx)	i2 = nx;
	if(j2>=ny)	j2 = ny;

	// debugging
	//CString TempStr;
	//TempStr.Format("x = (%d...%d), y = (%d...%d)", i1, i2, j1, j2);
	//AfxMessageBox(TempStr);

	bnx = i2 - i1;
	bny = j2 - j1;

	// prepare buffer for line
	char * pDiffBuf = (char*)calloc(TileSize * 4, 1);
	if(!pDiffBuf) {
		return ERROR_ALLOC_FAILED;
	}
	char * pOutBuf = (char*)calloc(TileSize * 4, 1);
	if(!pOutBuf) {
		return ERROR_ALLOC_FAILED;
	}

	// get min/max alt in block (used for vert scale)
	HFmin = 0;
	HFmax = 0;
	for(j=j1; j<j2; j++) {
		for(i=i1; i<i2; i++) {
			// find max diff in line

			f = pMapData[i + j * nx];
		
			if(j==j1 && i==i1) {
				HFmin = HFmax = f;
			} else {
				if(f<HFmin)	HFmin = f;
				if(f>HFmin)	HFmax = f;
			}
		}
	}

	// number of int levels required for this block
	float BlockLevels = (HFmax - HFmin) / Precis + 1;

	// debugging
	//TempStr.Format("Block %d,%d: %.1f", bi, bj, BlockLevels);
	//theAPI.ReportError(TempStr);

	// calc scale
	VertScale = (HFmax - HFmin)/(float)BlockLevels;
	VertOffset = HFmin;
	if(VertScale<=0)	VertScale = 1.0f; // this is for niceness

	// write the block scaling
	char ScaleBuf[8];
	memcpy(ScaleBuf, &VertScale, 4);
	memcpy(ScaleBuf+4, &VertOffset, 4);
	if(8!=hfzWrite(fs, ScaleBuf, 8)) {
		return ERROR_WRITE_FAILED;
	}

	// save block line-by-line
	for(j=j1; j<j2; j++) {
	
		// get first val
		f = pMapData[i1 + j * nx];
		FirstVal = (long)((f-VertOffset)/VertScale);
		int LastVal = FirstVal;

		// find max diff in line
		int Diff;
		int MaxDev = 0;
		for(i=i1+1; i<i2; i++) {

			// find max diff in line
			f = pMapData[i + j * nx];
		
			TempInt=(int)((f-VertOffset)/VertScale);
			Diff = TempInt - LastVal;

			int zi = i-i1-1;
			memcpy((char*)pDiffBuf+zi*4, &Diff, 4);

			LastVal = TempInt;

			MaxDev = MaxDev>abs(Diff)?MaxDev:abs(Diff);
		}

		// should we use 8, 16 or 32 bit pixels?
		char LineDepth = 4;
		if(MaxDev<=127) {
			LineDepth = 1;
		} else
		if(MaxDev<=32768) {
			LineDepth = 2;
		}

		// write line header
		char BlockBuf[5];
		memcpy(BlockBuf, &LineDepth, 1); // store line depth
		memcpy(BlockBuf+1, &FirstVal, 4); // store first value (32bit precis)
		if(5!=hfzWrite(fs, BlockBuf, 5)) {
			return ERROR_WRITE_FAILED;
		}
		
		// now write block
		char* pOutBuf2 = pOutBuf;
		for(i=i1+1; i<i2; i++) {

			int zi = i-i1-1;
			memcpy(&Diff, (char*)pDiffBuf+zi*4, 4);
			switch(LineDepth) {
			case 1:
				c = (char)Diff;
				memcpy(pOutBuf2, &c, LineDepth);
				break;
			case 2:
				s = (short)Diff;
				memcpy(pOutBuf2, &s, LineDepth);
				break;
			case 4:
				memcpy(pOutBuf2, &Diff, LineDepth);
				break;
			}

			pOutBuf2+=LineDepth;
		}

		int len = LineDepth * (i2-i1-1);
		if(len!=hfzWrite(fs, pOutBuf, len))
			return ERROR_WRITE_FAILED;
	}

	free(pDiffBuf);
	free(pOutBuf);

	return ERROR_OK;
}

//
//
// Low-level HFZ functions
//
//

hfzFile* hfzOpen(const char* lpFileName, long Format, const char* lpOptions) {
	
	hfzFile* fs = new hfzFile;

	switch(Format) {
	case FORMAT_HF2: 
		if(!(fs->raw = fopen(lpFileName, lpOptions))) {
			delete fs;
			return NULL;
		}
		break;
	case FORMAT_HF2_GZ: 
		if(!(fs->gz = gzopen(lpFileName, lpOptions))) {
			delete fs;
			return NULL;
		}
		break;
	default: 
		delete fs;
		return NULL;
	}

	fs->Format = Format;

	return fs;
}

void hfzClose(hfzFile* fs) {

	if(!fs)
		return;

	switch(fs->Format) {
	case FORMAT_HF2:	fclose(fs->raw); break;
	case FORMAT_HF2_GZ: gzclose(fs->gz); break;
	}

	delete fs;
}

long hfzRead(hfzFile* fs, void* pData, int len) {
	if(!fs)
		return ERROR_INVALID_HANDLE;

	if(!len)
		return 0;

	if(!pData)
		return ERROR_INVALID_HANDLE;

    switch(fs->Format) {
	case FORMAT_HF2:    return fread(pData, len, 1, fs->raw);
	case FORMAT_HF2_GZ: return gzread(fs->gz, pData, len);
	}

	return ERROR_INVALID_FORMAT;
}

long hfzWrite(hfzFile* fs, void* pData, int len) {
	if(!fs)
		return ERROR_INVALID_HANDLE;

	if(!len)
		return 0;

	if(!pData)
		return ERROR_INVALID_HANDLE;

	switch(fs->Format) {
	case FORMAT_HF2:    return fwrite(pData, 1, len, fs->raw);
	case FORMAT_HF2_GZ: return gzwrite(fs->gz, pData, len);
	}

	return ERROR_INVALID_FORMAT;
}

//
//
// HFZ helper functions
//
//


const char* hfzGetErrorStr(long ErrorCode) {
	switch(ErrorCode) {
	case ERROR_OPEN_FAILED:			return "Cannot open file";
	case ERROR_WRITE_FAILED:		return "Write failed";
	case ERROR_READ_FAILED:			return "Read failed";
	case ERROR_ALLOC_FAILED:		return "Cannot allocate buffer";
	case ERROR_INVALID_HANDLE:		return "Invalid handle";
	case ERROR_INVALID_FORMAT:		return "Invalid format ID";
	case ERROR_INVALID_PRECIS:		return "Invalid precision";
	case ERROR_INVALID_MAP_SIZE:	return "Invalid map size";
	case ERROR_INVALID_TILE_SIZE:	return "Invalid tile size";
	case ERROR_WRITE_HEADER_FAILED: return "Error writing header";
	case ERROR_WRITE_EXTHEAD_FAILED: return "Error writign extended header";
	}

	return NULL;
}

void hfzResetHeader(hfzHeader& fh) {
	if(fh.ExtHeaderLength) {
		fh.ExtHeaderLength = 0;
		if(fh.pExtHeaderData) free(fh.pExtHeaderData);
	}

	fh.FileVersionNo = 0;
	fh.nx = 0;
	fh.ny = 0;
	fh.HorizScale = 1;
	fh.Precis = 0.01;
	fh.TileSize = 256;
}